#!/usr/bin/env python3

import argparse
import json
import logging
import os
import sys
from datetime import datetime
from pathlib import Path
from typing import Iterator, Optional

import grpc
import numpy as np
import soundfile
from google.protobuf.duration_pb2 import Duration
from google.protobuf.json_format import MessageToDict
from phonexia.grpc.common.core_pb2 import (
    Audio,
    RawAudioConfig,
    TimeRange,
)
from phonexia.grpc.technologies.time_analysis.v1.time_analysis_pb2 import (
    AnalyzeRequest,
    AnalyzeResponse,
)
from phonexia.grpc.technologies.time_analysis.v1.time_analysis_pb2_grpc import (
    TimeAnalysisStub,
)


def print_json(message, output_file: Optional[Path] = None) -> None:
    json.dump(
        message, output_file.open("w") if output_file else sys.stdout, indent=2, ensure_ascii=False
    )


def message_to_dict(message):
    return MessageToDict(
        message,
        always_print_fields_with_no_presence=True,
        preserving_proto_field_name=True,
        use_integers_for_enums=False,
    )


def time_to_duration(time: float) -> Optional[Duration]:
    if time is None:
        return None
    duration = Duration()
    duration.seconds = int(time)
    duration.nanos = int((time - duration.seconds) * 1e9)
    return duration


def make_request(
    file: str,
    start: Optional[float],
    end: Optional[float],
    use_raw_audio: bool,
) -> Iterator[AnalyzeRequest]:
    time_range = TimeRange(start=time_to_duration(start), end=time_to_duration(end))
    chunk_size = 1024 * 100
    if use_raw_audio:
        with soundfile.SoundFile(file) as r:
            raw_audio_config = RawAudioConfig(
                channels=r.channels,
                sample_rate_hertz=r.samplerate,
                encoding=RawAudioConfig.AudioEncoding.PCM16,
            )
            for data in r.blocks(blocksize=r.samplerate, dtype="float32"):
                int16_info = np.iinfo(np.int16)
                data_scaled = np.clip(
                    data * (int16_info.max + 1), int16_info.min, int16_info.max
                ).astype("int16")
                yield AnalyzeRequest(
                    audio=Audio(
                        content=data_scaled.flatten().tobytes(),
                        raw_audio_config=raw_audio_config,
                        time_range=time_range,
                    )
                )
                time_range = None
                raw_audio_config = None

    else:
        with open(file, mode="rb") as fd:
            while chunk := fd.read(chunk_size):
                yield AnalyzeRequest(audio=Audio(content=chunk, time_range=time_range))
                time_range = None


def write_result(response: AnalyzeResponse, output_file: Optional[str] = None) -> None:
    logging.debug("Writing matches to {}".format(output_file if output_file else "'stdout'"))
    message = message_to_dict(message=response)
    print_json(message, output_file)


def analyze(
    channel: grpc.Channel,
    input_file: Path,
    output_file: Optional[Path],
    start: Optional[float],
    end: Optional[float],
    metadata: Optional[list],
    use_raw_audio: bool,
) -> None:
    logging.info(
        "Analyzing {input} -> {output}".format(
            input=input_file, output=(output_file if output_file else "'stdout'")
        )
    )

    response_it = make_request(
        file=input_file,
        start=start,
        end=end,
        use_raw_audio=use_raw_audio,
    )

    stub = TimeAnalysisStub(channel)
    for response in stub.Analyze(response_it, metadata=metadata):
        write_result(response, output_file)


# Main program
def handle_grpc_error(e: grpc.RpcError):
    logging.error(f"gRPC call failed with status code: {e.code()}")
    logging.error(f"Error details: {e.details()}")

    if e.code() == grpc.StatusCode.UNAVAILABLE:
        logging.error("Service is unavailable. Please try again later.")
    elif e.code() == grpc.StatusCode.INVALID_ARGUMENT:
        logging.error("Invalid arguments were provided to the RPC.")
    elif e.code() == grpc.StatusCode.DEADLINE_EXCEEDED:
        logging.error("The RPC deadline was exceeded.")
    else:
        logging.error(f"An unexpected error occurred: {e.code()} - {e.details()}")


def check_file_exists(path: Path) -> Path:
    if not os.path.isfile(path):
        raise argparse.ArgumentTypeError(f"File '{path}' does not exist.")
    return Path(path)


def main():
    parser = argparse.ArgumentParser(
        description=("Analyze time based traits in a multichannel audio file"),
    )
    parser.add_argument(
        "-H",
        "--host",
        type=str,
        default="localhost:8080",
    )
    parser.add_argument(
        "-l",
        "--log_level",
        type=str,
        default="error",
        choices=["critical", "error", "warning", "info", "debug"],
    )
    parser.add_argument(
        "--metadata",
        metavar="key=value",
        nargs="+",
        type=lambda x: tuple(x.split("=")),
        help="Custom client metadata",
    )
    parser.add_argument(
        "-i",
        "--input",
        type=check_file_exists,
        help="Multichannel input audio file.",
        required=True,
    )
    parser.add_argument(
        "-o",
        "--output",
        type=Path,
        default=None,
        help="Output result to a file. If omitted, output to stdout.",
    )
    parser.add_argument("--use_ssl", action="store_true", help="Use SSL connection")
    parser.add_argument("--start", type=float, help="Audio start time")
    parser.add_argument("--end", type=float, help="Audio end time")
    parser.add_argument("--use_raw_audio", action="store_true", help="Send a raw audio")

    args = parser.parse_args()

    output_file: Optional[Path] = Optional
    output_file = args.output

    if args.start is not None and args.start < 0:
        raise ValueError("Parameter 'start' must be a non-negative float.")

    if args.end is not None and args.end <= 0:
        raise ValueError("Parameter 'end' must be a positive float.")

    logging.basicConfig(
        level=args.log_level.upper(),
        format="[%(asctime)s.%(msecs)03d] [%(levelname)s] %(message)s",
        datefmt="%Y-%m-%d %H:%M:%S",
    )

    try:
        logging.info(f"Connecting to {args.host}")
        channel = (
            grpc.secure_channel(target=args.host, credentials=grpc.ssl_channel_credentials())
            if args.use_ssl
            else grpc.insecure_channel(target=args.host)
        )

        start_time = datetime.now()

        analyze(
            channel=channel,
            input_file=args.input,
            output_file=output_file,
            start=args.start,
            end=args.end,
            metadata=args.metadata,
            use_raw_audio=args.use_raw_audio,
        )

        logging.debug(f"Elapsed time {(datetime.now() - start_time)}")

    except grpc.RpcError as e:
        handle_grpc_error(e)
        exit(1)
    except Exception:
        logging.exception("Unknown error")
        exit(1)


if __name__ == "__main__":
    main()
